package com.huawei.dli.flink.demo.sink;

import java.security.PrivilegedExceptionAction;
import java.sql.Connection;
import java.sql.DriverManager;

import com.huawei.dli.flink.demo.utils.LoginUtil;
import com.huawei.dli.flink.demo.utils.HiveClientInfo;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.runtime.execution.SuppressRestartsException;
import org.apache.flink.streaming.api.functions.sink.RichSinkFunction;
import org.apache.hadoop.security.UserGroupInformation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static org.apache.hadoop.fs.CommonConfigurationKeysPublic.HADOOP_SECURITY_AUTHENTICATION;
import static org.apache.hadoop.fs.CommonConfigurationKeysPublic.HADOOP_SECURITY_AUTHORIZATION;

public class HiveSink<IN> extends RichSinkFunction<IN> {
  private static final Logger logger = LoggerFactory.getLogger(HiveSink.class);

  private String tmpDir = "/tmp/hiveSink.";
  private String userPrincipal;
  private String zkServerPrincipal;

  private String dbName = "";

  private transient HiveClientInfo clientInfo;
  private boolean isSecurityMode = false;

  transient Connection conn;

  public HiveSink(String userPrincipal, String zkServerPrincipal, String dbName) {
    this.userPrincipal = userPrincipal;
    this.zkServerPrincipal = zkServerPrincipal;
    this.dbName = dbName;
  }

  @Override
  public void open(Configuration parameters) throws Exception {
    clientInfo = new HiveClientInfo("hiveclient.properties");
    isSecurityMode = "KERBEROS".equalsIgnoreCase(clientInfo.getAuth());
    if (isSecurityMode) {
      kerberosLogin();
    }

    super.open(parameters);
    connectionDatabase();
  }

  @Override
  public void close() throws Exception {
    if (conn != null) {
      conn.close();
    }

    try {
      LoginUtil.deleteFile(tmpDir, "user.keytab");
      LoginUtil.deleteFile(tmpDir, "krb5.conf");
    } catch (Exception e) {
      logger.error("Failed to close source.", e);
    }

    super.close();
  }

  /**
   * Config for kerberos auth hive
   */
  private void kerberosLogin() throws Exception {
    org.apache.hadoop.conf.Configuration conf = new org.apache.hadoop.conf.Configuration();
    conf.set(HADOOP_SECURITY_AUTHENTICATION, "kerberos");
    conf.set(HADOOP_SECURITY_AUTHORIZATION, "true");

    String userKeytabFile = LoginUtil.createFile(tmpDir, "user.keytab", this.getClass());
    String krb5File = LoginUtil.createFile(tmpDir, "krb5.conf", this.getClass());
    LoginUtil.setJaasConf(LoginUtil.Module.ZOOKEEPER.getName(), userPrincipal, userKeytabFile);
    LoginUtil.setZookeeperServerPrincipal(zkServerPrincipal);
    LoginUtil.login(userPrincipal, userKeytabFile, krb5File, conf);
  }

  private void connectionDatabase() {
    try {
      Class.forName("org.apache.hive.jdbc.HiveDriver");

      StringBuilder hiveUrlBuilder = new StringBuilder("jdbc:hive2://").append(clientInfo.getZkQuorum())
        .append("/").append(dbName);
      if (isSecurityMode) {
        hiveUrlBuilder.append(";serviceDiscoveryMode=")
          .append(clientInfo.getServiceDiscoveryMode())
          .append(";zooKeeperNamespace=")
          .append(clientInfo.getZooKeeperNamespace())
          .append(";sasl.qop=")
          .append(clientInfo.getSaslQop())
          .append(";auth=")
          .append(clientInfo.getAuth())
          .append(";principal=")
          .append(clientInfo.getPrincipal())
          .append(";");
      } else {
        hiveUrlBuilder.append(";serviceDiscoveryMode=")
          .append(clientInfo.getServiceDiscoveryMode())
          .append(";zooKeeperNamespace=")
          .append(clientInfo.getZooKeeperNamespace())
          .append(";auth=none");
      }

      logger.info("hiveUrl: " + hiveUrlBuilder.toString());
      UserGroupInformation.getLoginUser().doAs(new PrivilegedExceptionAction<Void>() {
        public Void run() throws Exception {
          boolean isSuccess = false;
          int i = 3;
          while (!isSuccess && i > 0) {
            try {
              conn = DriverManager.getConnection(hiveUrlBuilder.toString(), "", "");
              logger.info("connect success");
              isSuccess = true;
            } catch (Exception e) {
              logger.error("connect failed", e);
              i--;
            }
          }
          return null;
        }
      });
    } catch (Exception ex) {
      logger.error("Failed to connect to hive", ex);
      throw new SuppressRestartsException(ex);
    }
  }
}
